/*
 * system.h - system hardware device configuration values
 * Part of TinyG project
 *
 * Copyright (c) 2010 - 2012 Alden S. Hart Jr.
 *
 * This file ("the software") is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2 as published by the
 * Free Software Foundation. You should have received a copy of the GNU General Public
 * License, version 2 along with the software.  If not, see <http://www.gnu.org/licenses/>.
 *
 * As a special exception, you may use this file as part of a software library without
 * restriction. Specifically, if other files instantiate templates or use macros or
 * inline functions from this file, or you compile this file and link it with  other
 * files to produce an executable, this file does not by itself cause the resulting
 * executable to be covered by the GNU General Public License. This exception does not
 * however invalidate any other reasons why the executable file might be covered by the
 * GNU General Public License.
 *
 * THE SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT WITHOUT ANY
 * WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
 * SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/*
 * INTERRUPT USAGE - TinyG uses a lot of them all over the place
 *
 *  HI  Stepper DDA pulse generation        (set in stepper.h)
 *  HI  Stepper load routine SW interrupt   (set in stepper.h)
 *  HI  Dwell timer counter                 (set in stepper.h)
 *  LO  Segment execution SW interrupt      (set in stepper.h)
 *  MED GPIO1 switch port                   (set in gpio.h)
 *  MED Serial RX for USB & RS-485          (set in xio_usart.h)
 *  LO  Serial TX for USB & RS-485          (set in xio_usart.h)
 *  LO  Real time clock interrupt           (set in xmega_rtc.h)
 */
#ifndef system_h
#define system_h

#define STEPCRAFT                               // Modifications by Thomas Keil to use with Stepcraft machines

void sys_init(void);                    // master hardware init
void sys_port_bindings(float hw_version);
void sys_get_id(char *id);

#define SYS_ID_LEN 12                   // length of system ID string from sys_get_id()

/* CPU clock */

#undef F_CPU                            // set for delays
#define F_CPU 32000000UL                // should always precede <avr/delay.h>

// Clock Crystal Config. Pick one:
//#define __CLOCK_INTERNAL_32MHZ TRUE   // use internal oscillator
#define __CLOCK_EXTERNAL_8MHZ   TRUE    // uses PLL to provide 32 MHz system clock
//#define __CLOCK_EXTERNAL_16MHZ TRUE       // uses PLL to provide 32 MHz system clock

/*** Motor, output bit & switch port assignments ***
 *** These are not all the same, and must line up in multiple places in gpio.h ***
 * Sorry if this is confusing - it's a board routing issue
 */
#ifdef STEPCRAFT
#define PORT_MOTOR      PORTB           // all motor signals resides at one port
#else
#define PORT_MOTOR_1    PORTA           // motors mapped to ports
#define PORT_MOTOR_2    PORTF
#define PORT_MOTOR_3    PORTE
#define PORT_MOTOR_4    PORTD
#endif

#ifdef STEPCRAFT
#define PORT_SWITCH     PORTA           // switches at one port
#else
#define PORT_SWITCH_X   PORTA           // Switch axes mapped to ports
#define PORT_SWITCH_Y   PORTD
#define PORT_SWITCH_Z   PORTE
#define PORT_SWITCH_A   PORTF
#endif

#ifdef STEPCRAFT
#define PORT_OUT        PORTC           // C.0 Spindle
#else
#define PORT_OUT_V7_X   PORTA           // v7 mapping - Output bits mapped to ports
#define PORT_OUT_V7_Y   PORTF
#define PORT_OUT_V7_Z   PORTD
#define PORT_OUT_V7_A   PORTE

#define PORT_OUT_V6_X   PORTA           // v6 and earlier mapping - Output bits mapped to ports
#define PORT_OUT_V6_Y   PORTF
#define PORT_OUT_V6_Z   PORTE
#define PORT_OUT_V6_A   PORTD
#endif

// These next four must be changed when the PORT_MOTOR_* definitions change!
#ifdef STEPCRAFT
#define PORTCFG_VP0MAP_PORT_MOTOR_1_gc PORTCFG_VP0MAP_PORTB_gc  // this virtual port is used for motor control
#define PORTCFG_VP1MAP_PORT_MOTOR_2_gc PORTCFG_VP1MAP_PORTF_gc  // the are not used
#define PORTCFG_VP2MAP_PORT_MOTOR_3_gc PORTCFG_VP2MAP_PORTE_gc
#define PORTCFG_VP3MAP_PORT_MOTOR_4_gc PORTCFG_VP3MAP_PORTD_gc
#else
#define PORTCFG_VP0MAP_PORT_MOTOR_1_gc PORTCFG_VP0MAP_PORTA_gc
#define PORTCFG_VP1MAP_PORT_MOTOR_2_gc PORTCFG_VP1MAP_PORTF_gc
#define PORTCFG_VP2MAP_PORT_MOTOR_3_gc PORTCFG_VP2MAP_PORTE_gc
#define PORTCFG_VP3MAP_PORT_MOTOR_4_gc PORTCFG_VP3MAP_PORTD_gc
#endif

#define PORT_MOTOR_1_VPORT  VPORT0              // Stepcraft: the port for every motor mapped to PortB
#ifndef PORT_MOTOR
#define PORT_MOTOR_2_VPORT  VPORT1              // not used for Stepcraft
#define PORT_MOTOR_3_VPORT  VPORT2
#define PORT_MOTOR_4_VPORT  VPORT3
#endif

/*
 * Port setup - Stepper / Switch Ports:
 *  b0  (out) step          (SET is step,  CLR is rest)
 *  b1  (out) direction     (CLR = Clockwise)
 *  b2  (out) motor enable  (CLR = Enabled)
 *  b3  (out) microstep 0
 *  b4  (out) microstep 1
 *  b5  (out) output bit for GPIO port1
 *  b6  (in) min limit switch on GPIO 2 (note: motor controls and GPIO2 port mappings are not the same)
 *  b7  (in) max limit switch on GPIO 2 (note: motor controls and GPIO2 port mappings are not the same)
 */

#ifdef PORT_MOTOR
#define MOTOR_PORT_DIR_gm 0xFF  // dir settings: all motors (DIR/CLK) connected to one port
#else
#define MOTOR_PORT_DIR_gm 0x3F  // dir settings: lower 6 out, upper 2 in
#endif

enum cfgPortBits {          // motor control port bit positions
    STEP_BIT_bp = 0,        // bit 0
    DIRECTION_BIT_bp,       // bit 1
    MOTOR_ENABLE_BIT_bp,    // bit 2
    MICROSTEP_BIT_0_bp,     // bit 3
    MICROSTEP_BIT_1_bp,     // bit 4
    GPIO1_OUT_BIT_bp,       // bit 5 (4 gpio1 output bits; 1 from each axis)
    SW_MIN_BIT_bp,          // bit 6 (4 input bits for homing/limit switches)
    SW_MAX_BIT_bp           // bit 7 (4 input bits for homing/limit switches)
};

#ifdef PORT_MOTOR
#define STEP_SET_1          (PORT_MOTOR_1_VPORT.OUT |= (1<<6))  // turn step bit on
#define STEP_CLR_1          (PORT_MOTOR_1_VPORT.OUT &= ~(1<<6))
#define STEP_SET_2          (PORT_MOTOR_1_VPORT.OUT |= (1<<4))
#define STEP_CLR_2          (PORT_MOTOR_1_VPORT.OUT &= ~(1<<4))
#define STEP_SET_3          (PORT_MOTOR_1_VPORT.OUT |= (1<<2))
#define STEP_CLR_3          (PORT_MOTOR_1_VPORT.OUT &= ~(1<<2))
#define STEP_SET_4          (PORT_MOTOR_1_VPORT.OUT |= (1<<0))
#define STEP_CLR_4          (PORT_MOTOR_1_VPORT.OUT &= ~(1<<0))
#define CW_1                (PORT_MOTOR_1_VPORT.OUT &= ~(1<<7))
#define CCW_1               (PORT_MOTOR_1_VPORT.OUT |= (1<<7))
#define CW_2                (PORT_MOTOR_1_VPORT.OUT &= ~(1<<5))
#define CCW_2               (PORT_MOTOR_1_VPORT.OUT |= (1<<5))
#define CW_3                (PORT_MOTOR_1_VPORT.OUT &= ~(1<<3))
#define CCW_3               (PORT_MOTOR_1_VPORT.OUT |= (1<<3))
#define CW_4                (PORT_MOTOR_1_VPORT.OUT &= ~(1<<1))
#define CCW_4               (PORT_MOTOR_1_VPORT.OUT |= (1<<1))
#define ENABLE_1
#define ENABLE_2
#define ENABLE_3
#define ENABLE_4
#define DISABLE_1
#define DISABLE_2
#define DISABLE_3
#define DISABLE_4
#else
#define STEP_SET_1          (PORT_MOTOR_1_VPORT.OUT |= STEP_BIT_bm)  // turn step bit on
#define STEP_CLR_1          (PORT_MOTOR_1_VPORT.OUT &= ~STEP_BIT_bm) // turn step bit off in ~1 uSec
#define STEP_SET_2          (PORT_MOTOR_2_VPORT.OUT |= STEP_BIT_bm)  // turn step bit on
#define STEP_CLR_2          (PORT_MOTOR_2_VPORT.OUT &= ~STEP_BIT_bm) // turn step bit off in ~1 uSec
#define STEP_SET_3          (PORT_MOTOR_3_VPORT.OUT |= STEP_BIT_bm)  // turn step bit on
#define STEP_CLR_3          (PORT_MOTOR_3_VPORT.OUT &= ~STEP_BIT_bm) // turn step bit off in ~1 uSec
#define STEP_SET_4          (PORT_MOTOR_4_VPORT.OUT |= STEP_BIT_bm)  // turn step bit on
#define STEP_CLR_4          (PORT_MOTOR_4_VPORT.OUT &= ~STEP_BIT_bm) // turn step bit off in ~1 uSec
#define CW_1                (PORT_MOTOR_1_VPORT.OUT &= ~DIRECTION_BIT_bm)
#define CCW_1               (PORT_MOTOR_1_VPORT.OUT |= DIRECTION_BIT_bm)
#define CW_2                (PORT_MOTOR_2_VPORT.OUT &= ~DIRECTION_BIT_bm)
#define CCW_2               (PORT_MOTOR_2_VPORT.OUT |= DIRECTION_BIT_bm)
#define CW_3                (PORT_MOTOR_3_VPORT.OUT &= ~DIRECTION_BIT_bm)
#define CCW_3               (PORT_MOTOR_3_VPORT.OUT |= DIRECTION_BIT_bm)
#define CW_4                (PORT_MOTOR_4_VPORT.OUT &= ~DIRECTION_BIT_bm)
#define CCW_4               (PORT_MOTOR_4_VPORT.OUT |= DIRECTION_BIT_bm)
#define ENABLE_1            (PORT_MOTOR_1_VPORT.OUT &= ~MOTOR_ENABLE_BIT_bm)
#define ENABLE_2            (PORT_MOTOR_2_VPORT.OUT &= ~MOTOR_ENABLE_BIT_bm)
#define ENABLE_3            (PORT_MOTOR_3_VPORT.OUT &= ~MOTOR_ENABLE_BIT_bm)
#define ENABLE_4            (PORT_MOTOR_4_VPORT.OUT &= ~MOTOR_ENABLE_BIT_bm)
#define DISABLE_1           (PORT_MOTOR_1_VPORT.OUT |= MOTOR_ENABLE_BIT_bm)
#define DISABLE_2           (PORT_MOTOR_2_VPORT.OUT |= MOTOR_ENABLE_BIT_bm)
#define DISABLE_3           (PORT_MOTOR_3_VPORT.OUT |= MOTOR_ENABLE_BIT_bm)
#define DISABLE_4           (PORT_MOTOR_4_VPORT.OUT |= MOTOR_ENABLE_BIT_bm)
#endif

#define STEP_BIT_bm         (1<<STEP_BIT_bp)
#define DIRECTION_BIT_bm    (1<<DIRECTION_BIT_bp)
#define MOTOR_ENABLE_BIT_bm (1<<MOTOR_ENABLE_BIT_bp)
#define MICROSTEP_BIT_0_bm  (1<<MICROSTEP_BIT_0_bp)
#define MICROSTEP_BIT_1_bm  (1<<MICROSTEP_BIT_1_bp)
#define GPIO1_OUT_BIT_bm    (1<<GPIO1_OUT_BIT_bp)   // spindle and coolant output bits
#define SW_MIN_BIT_bm       (1<<SW_MIN_BIT_bp)      // minimum switch inputs
#define SW_MAX_BIT_bm       (1<<SW_MAX_BIT_bp)      // maximum switch inputs

/* Bit assignments for GPIO1_OUTs for spindle, PWM and coolant */

#define SPINDLE_BIT         0x08        // spindle on/off
#define SPINDLE_DIR         0x04        // spindle direction, 1=CW, 0=CCW
#define SPINDLE_PWM         0x02        // spindle PWMs output bit
#define MIST_COOLANT_BIT    0x01        // coolant on/off - these are the same due to limited ports
#define FLOOD_COOLANT_BIT   0x10        // coolant on/off

#define SPINDLE_LED         0
#define SPINDLE_DIR_LED     1
#define SPINDLE_PWM_LED     2
#define COOLANT_LED         3

#define INDICATOR_LED       SPINDLE_DIR_LED // can use the spindle direction as an indicator LED

/* Timer assignments - see specific modules for details) */

#define TIMER_DDA           TCC0        // DDA timer    (see stepper.h)
#define TIMER_DWELL         TCD0        // Dwell timer  (see stepper.h)
#define TIMER_LOAD          TCE0        // Loader timer (see stepper.h)
#define TIMER_EXEC          TCF0        // Exec timer   (see stepper.h)
#define TIMER_5             TCC1        // unallocated timer
#define TIMER_PWM1          TCD1        // PWM timer #1 (see pwm.c)
#define TIMER_PWM2          TCE1        // PWM timer #2 (see pwm.c)


/**** Device singleton - global structure to allow iteration through similar devices ****/
/*
    Ports are shared between steppers and GPIO so we need a global struct.
    Each xmega port has 3 bindings; motors, switches and the output bit

    The initialization sequence is important. the order is:
        - sys_init()    binds all ports to the device struct
        - st_init()     sets IO directions and sets stepper VPORTS and stepper specific functions
        - gpio_init()   sets up input and output functions and required interrupts

    Care needs to be taken in routines that use ports not to write to bits that are
    not assigned to the designated function - ur unpredicatable results will occur
*/

typedef struct deviceSingleton {

#ifdef PORT_MOTOR
    PORT_t *st_port;            // bindings for single stepper motor port (stepper.c)
#else
    PORT_t *st_port[MOTORS];    // bindings for stepper motor ports (stepper.c)
#endif

#ifdef PORT_SWITCH
    PORT_t *sw_port;    // bindings for switch ports (GPIO2)
#else
    PORT_t *sw_port[MOTORS];    // bindings for switch ports (GPIO2)
#endif

#ifdef PORT_OUT
    PORT_t *out_port;
#else
    PORT_t *out_port[MOTORS];   // bindings for output ports (GPIO1)
#endif
} deviceSingleton_t;
deviceSingleton_t device;

#endif
